# Copyright 2018 miruka
# This file is part of lunafind, licensed under LGPLv3.

"Search, filter, inspect, download posts from Danbooru and local folders."

__project_name__ = "lunakit"
__pkg_name__     = "lunafind"
__version__      = "0.5.4-2"
__status__       = "Development"
# __status__     = "Production"

__author__ = "miruka"
__email__  = "miruka@disroot.org"

__license__ = "LGPLv3"
